(* =============================================================================
================ FONCTIONS GENERALES SUR LES LISTES ============================
===============================================================================*)



let suppression p l =  
  (* rsultat : la liste l prive de ses lments qui vrifient la proprit p *)
  List.rev (List.fold_left (fun l1 x -> if p x then l1 else x::l1) [] l);;

let enlever x l = 
  (* rsultat : la liste l prive des occurrences de x *)
  suppression (fun y -> x = y) l;;

let rec disjoint l1 l2 = 
  (* rsultat : true ssi les 2 listes n'ont pas d'lments commun *)
  match l1 with
    | [] -> true
    | dl1::fl1 -> not (List.mem dl1 l2) && (disjoint fl1 l2);;


  

let rec ens_de_liste l = 
  (* rsultat : la liste l avec suppression de ses lments rpts  *)
  match l with
    | [] -> []
    | t::f -> if List.mem t f then ens_de_liste f else t::(ens_de_liste f);;


let red infe x l1 =
  (* red infe  x l1 = (a, l2) o
   a vaut true si et seulement si l1 comporte un lment strictement infrieur  x 
   l2 est la liste obtenue en enlevant de l1 les lments suprieurs ou gaux  x *)
  let rec aux = function
    | [] -> (false, [])
    | y::l2 -> let (a,l3) = aux l2 in 
      if (infe x y) then (a, l3) else if (infe y x) then (true,y::l3) else (a,y::l3)
  in aux l1;;
      

let rduction infe l1 =
  (* calcule la liste des lments minimaux de l1 pour l'ordre partiel infe *)
  let rec aux = function
    | [] -> []
    | x::l2 -> let (a, l3) = red infe x l2 in
      if a then aux l3 else x::aux l3
  in aux l1;;


let rec sous_ens e =
  (* liste des sous-ensembles de la liste e *)
  match e with
    | [] -> [[]]
    | t::f -> let sef = sous_ens f in
	(List.map (fun x -> t::x) sef)@ sef;;

let sous_ens2 e =
  (* liste des sous-ensembles de e ayant au moins deux lments *)
  suppression (fun x -> List.length x < 2) (sous_ens e);;


(* FIN FONCTIONS GENERALES LISTES *)


(* ===========================================================================================
=======================  RESOLUTION     ======================================================
============================================================================================*)

open Types;;
	       
let oppose l1 l2 = 
  (* rsultat :  true si et seulement si l1 et l2 sont des littraux opposs *)
  match (l1, l2) with
    | (Pos p1, Neg p2) -> p1 = p2
    | (Neg p1, Pos p2) -> p1 = p2
    | _ -> false ;;

let opp = 
  (* rsultat : le littral oppos  l *)
  function 
    | (Pos x) -> Neg x
    | (Neg x) -> Pos x
;;

let var l = 
  (* rsultat :  si l est un littral positif alors l sinon l'oppos de l *)
    match l with
    | (Pos p1) as l1 -> l1
    | (Neg p1) -> Pos p1 ;;


let rec valide e = 
  (* rsultat : true ssi e est une liste valide de littraux *)
  match e with
   | [] -> false 
   | t::f -> (List.mem (opp t) f) || valide f;;

let valide_clause c = 
  (*rsultat : true ssi c est une clause valide  *)
  valide c.corps ;;

let suppression_valides  l = 
  (* enlve de l les clauses  valides *)
  suppression valide_clause l;;



(* =======================================================================================
==================================== CALCUL DU RESOLVANT (1 ordre)  =====================
======================================================================================= *)


(* ================================== SUBSTITUTION (1 ordre) ========================= *)

(* Une substitution est une liste de paires string * terme 
   Une quation est un couple de termes ou de littraux *)

let subst_terme x t e =
  (* remplace (Var x) par le terme t dans le terme e *)
  let rec subst = function
    |((Var y) as u) -> if x = y then t else u
    | (Comp (y,ly)) -> Comp (y, List.map subst ly)
  in subst e;;

let subst_eqs x t le = 
  (* remplace (Var x) par t dans toutes les quations entre termes de la liste le *) 
  List.map (fun (u,v) ->(subst_terme x t u),(subst_terme x t v)) le;;

let subst_res x t lr = 
  (* remplace (Var x) par t dans les termes de la substitution lr *)
  List.map (fun (u,v) -> u, (subst_terme x t v)) lr ;;

let subst_t s t =
  (* applique la substitution s au terme t *)
  let rec subst_t_aux = function
    |((Var x) as y) -> (try (List.assoc x s) with Not_found -> y)
    | (Comp(x,lx)) -> Comp(x, List.map subst_t_aux lx)
  in subst_t_aux t;;
	
let subst_l s x =
  (* applique la substitution s au littral x *)
  match x with
    | (Pos x) -> (Pos (subst_t s x))
    | (Neg x) -> (Neg (subst_t s x)) ;;

(* ================================= FIN SUBSTITUTION ==================================== *)
			


(* ================================== UNIFICATION (1 ordre) ============================= *)

exception Non_unifiable;;

let occur v t = 
  (* lve l'exception Non_unifiable si v est une variable du terme t *)
  let rec aux = function
    | (Var v') -> if v = v' then raise Non_unifiable
    | (Comp (_,la)) -> List.iter aux la
  in aux t;;	

let rec trans e  s  = 
  (* calcule par une mthode assez lente l'instance principale s' de la substitution s 
     o s' est solution du systme e d'quations entre termes 
     si s' n'existe pas, l'exception Non_unifiable est leve *)
  if e = [] then s
  else let (t1,t2)::le = e in
    if t1 = t2 then trans le s
    else
      match (t1,t2) with
	| (Var x,  t) -> occur x t; 
	    trans (subst_eqs x t le) (let sp =(subst_res x t s) in if List.mem (x,t) sp then sp else (x,t)::sp) 
	| (t, Var x) -> occur x t; 
	    trans (subst_eqs x t le) (let sp =(subst_res x t s) in if List.mem (x,t) sp then sp else (x,t)::sp)
	|  (Comp(x,lx),Comp(y,ly)) -> 
	     if x = y & (List.length lx) = (List.length ly)
	     then (trans ((List.combine lx ly) @ le)  s) 
	     else raise Non_unifiable ;;	
	

let unift t1 t2 = 
  (* si les termes t1 et t2 sont unifiables, cette fonction en
     calcule l'unificateur principal, sinon elle lve l'exception Non_unifiable *)
  trans [t1,t2] [];;	

let rec quations_de_liste l =
  (* transforme la liste [x1; x2; x3; ...] en une liste de couples [x1,x2; x2,x3; ...]*)
  match l with
    | [] -> []
    | [_] -> []
    | x1::x2::f -> (x1,x2)::quations_de_liste (x2::f);;

let quation_de_termes_de_quations_de_litterals e =
  (* change une quation entre littrals en une quation entre termes
     lve l'exception Non_unifiable s'il n'y a pas de solution *)
  match e with
    | Pos t1, Pos t2 -> t1,t2
    | Neg t1, Neg t2 -> t1,t2
    | _ -> raise Non_unifiable;;

let unifl l =
  (* si la liste l de littraux est unifiable, en calcule l'unificateur principal
     sinon lve l'exception Non_unifiable *)
  let l1 = quations_de_liste l in
  let l2 = List.map quation_de_termes_de_quations_de_litterals l1 in
    trans l2 [];;

(* ============================== FIN UNIFICATION ========================================== *)



(* ============================ FACTEURS D'UNE CLAUSE (1 ordre) =========================== *)

let facteurs c =
  (* rsultat : la liste des facteurs de c, c compris *)
  let l1 = sous_ens2 c.corps in
    (* l1 est la liste des listes qui sont des sous-listes de la liste des littraux de c 
       et qui ont au moins  2 lments *)
    c :: List.fold_left 
      (fun a b -> 
	 try let s = unifl b in 
	   {numero = 0; origine = Facteur (c,s); 
	    corps = ens_de_liste (List.map (subst_l s) c.corps)}::a
	 with Non_unifiable -> a) [] l1;;
	     
(* ================================ FIN FACTEURS =========================================== *)
    	


(* ============================ RENOMMAGE D'UNE CLAUSE (1 ordre) ========================== *)

let nv l =
  (* produit une chane "v{chiffre}" non lment de la liste l *)
  let rec aux i = let s = "v"^(string_of_int i) in 
    if (List.mem s l) then aux (i+1) else s
  in aux 0;;

let rec renommage e f = 
  (* e est une liste de chanes sans rptition, 
     la fonction calcule un renommage de domaine e par des variables
     qui ne figurent pas dans f *)
  match e with
    | [] -> []
    | t::fin -> 
	if List.mem t f 
	then let x = nv f in (t, Var x ):: (renommage fin (x::f))
	else renommage  fin f ;;

let rec vars_terme t =
  (* liste des variables du terme t *)
  match t with
    | Var x -> [x]
    | Comp(_,l) -> List.fold_left (fun a b -> a @ (vars_terme b)   ) [] l;;

let vars_litteral l =
  (* liste des variables du littral l *)
  match l with
    | Pos t -> vars_terme t
    | Neg t -> vars_terme t;;

let vars_litteraux c =
  (* liste sans rptition des variables de la liste c de litteraux *)
  ens_de_liste (List.fold_left (fun a b -> a @ (vars_litteral b)   ) [] c);;



(* ============================== FIN RENOMMAGE ===================================== *)



(* ============================ RESOLVANT BINAIRE (1 ordre) ======================== *)

let res_aux c1 l1 c2 l2 =
  (* l1 est un littral de c1, l2 est un littral de c2,
     rsultat : 
     - si l1 et (opp l2) ont s comme unificateur principal , alors [r]
     o r est le rsolvant de corps  s((c1-{l1})U(c2-{l2}))
     - si l1 et (opp l2) ne sont pas unifiables, alors []
  *)
  try let s = unifl [l1; opp l2] in
  let cc = (enlever l1 c1.corps)@(enlever l2 c2.corps) in
  let cr = List.map (subst_l s) cc in
    [{numero = 0; origine = ResBin(c1,c2,s); corps = ens_de_liste cr}] 
  with Non_unifiable -> [];;

let res_bin c1 c2 =
  (* rsultat : liste des rsolvants binaires des clauses c1 et c2  *)
  List.fold_left
    (fun lr l1 ->  
       lr @ 
       (List.fold_left 
	  (fun lr l2 -> lr @ (res_aux c1 l1 c2 l2)) [] c2.corps)) [] c1.corps;;






let rsolvant_sans_copie c1 c2 =
  (* pr : les clauses c1 et c2 n'ont pas de variables communes
     rsultat : la liste de tous les rsolvants de c1 et c2 *)
  (* on construit tous les facteurs de c1 et de c2 et on calcule les rsolvants binaires
     entre ces facteurs *)
  let lf1 = facteurs c1 and lf2 = facteurs c2 in
    List.fold_left
      (fun lr c ->  
	 lr @ 
	 (List.fold_left 
	    (fun lr d -> lr @ (res_bin c d)) [] lf2)) [] lf1;;





let copie c r = 
  (* construit une copie de la clause c avec le renommage r *)
  {numero = 0; origine = Copie (c,r); corps = List.map (subst_l r) c.corps}
	
    


(* dfinition de restrictions de la rsolution : au moins un des parents doit tre une
clause positive, au moins un des parents doit tre une clause ngative *)

let type_rsolution = ref 1;;
 
let pos = 1;;

let neg = -1;;

let nor = 0;;

let positive c =
  (* rsultat : true si  c est une clause positive *)
  List.for_all (function | Pos _ -> true | _ -> false) c.corps ;;

let ngative c = 
  (* rsultat : true si  c est une clause ngative *)
  List.for_all (function | Neg _ -> true | _ -> false) c.corps ;;

let rsolution_autorise t c1 c2 =
  (* rsultat : true si la rsolution de type t est autorise entre c1 et c2 *)
  t = neg & ((ngative c1) or (ngative c2))
    or t = pos & ((positive c1) or (positive c2))
    or t = nor;;

(* calcul des rsolvants *)

let rsolvants c1 c2 =
  (* rsultat : la liste des rsolvants des clauses  c1 et c2 *)
  if rsolution_autorise !type_rsolution c1 c2
  then
    begin
      (* copie de c2 si necessaire *)
      let vc1 = vars_litteraux c1.corps
      and vc2 = vars_litteraux c2.corps in
	if disjoint vc1 vc2 then rsolvant_sans_copie c1 c2
	else 
	  let c2copie = copie c2 (renommage vc2 vc1) in
	    rsolvant_sans_copie c1 c2copie
    end
  else [];;


(* restriction incomplte de la rsolution : contrle de la longueur des clauses et de la hauteur des clauses
on supprime toutes les clauses dpassant ces bornes *)



let rec hauteur_terme t = 
  (* rsultat : hauteur d'un terme t *)
  match t with 
    | Var _ -> 0
    | Comp (_,[]) -> 0
    | Comp (_,lt) -> 1+ (List.fold_left (fun m1 t -> max (hauteur_terme t) m1) 0 lt);;

let hauteur_litteral l =
  (* resultat : hauteur d'un littral *)
  match l with
    | Pos t -> 1 + hauteur_terme t
    | Neg t -> 1 + hauteur_terme t;;

let hauteur_clause c =
  (* rsultat : hauteur d'une clause c *)
  List.fold_left (fun m l -> max (hauteur_litteral l) m) 0 c.corps;;
  




(* ===================================== FIN RESOLVANT =============================== *)



(* ====================================================================================
 ======== SUBSOMPTION ENTRE CLAUSES (1 ordre) =========================================
 ======================================================================================= *)



(* ===================================== FILTRAGE (1 ordre) ========================== *)

exception Echec_filtre ;;

let rec filtre_aux s u v =
  (* s est une substitution, u et v sont des termes, 
     rsultat : une substitution s' extension de s tel que s'(u) = v 
     lve l'exception Echec_filtre si cette substitution  n'existe pas *)
  match u with
    | Var x -> (try if List.assoc x s = v then s else raise Echec_filtre
		with Not_found -> (x,v)::s)
    | Comp (f,lf) -> (match v with
		       | Var _ -> raise Echec_filtre
		       | Comp (g,lg) -> if f = g then filtre_aux_liste s lf lg else raise Echec_filtre)
and filtre_aux_liste s lf lg =
  match lf,lg with
      [],[] -> s
    | tf::ff, tg::fg -> filtre_aux_liste (filtre_aux s tf tg) ff fg
    | _ -> raise Echec_filtre;;

let filtre_termes t1 t2 =
  (* t1 et t2 sont des termes
     rsultat : une substitution s telle que s(t1) = t2 
     lve l'exception Echec_filtre si cette substitution n'existe pas *)
  filtre_aux [] t1 t2;;

let filtre_litteraux l1 l2 =
  (* l1 et l2 sont des littraux, 
     rsultat : une substitution s tel que s(l1) = l2 
     lve l'exception Echec_filtre si cette subsitution n'existe pas *)
  match l1,l2 with
    | Pos t1, Pos t2 -> filtre_termes t1 t2
    | Neg t1, Neg t2 -> filtre_termes t1 t2
    | _ -> raise Echec_filtre;;

(* =======================================  FIN FILTRAGE ============================ *)

let rec geler_terme t = 
  (* t est un terme
     rsultat : le terme dans lequel toute variable (Var x) est remplace par la constante (Comp (x,[]))
  *)
  match t with
    | Var x -> Comp (x,[])
    | Comp(f,lt) -> Comp(f,List.map geler_terme lt);;

let geler_litteral l = 
  (* l est un littral
     rsultat : le littral dans lequel toute variable (Var x) est remplace par la constante (Comp (x,[]))
  *)
  match l with
    | Pos t -> Pos (geler_terme t)
    | Neg t -> Neg (geler_terme t);;





let rec filtre c1 c2 =
  (* c1 et c2 sont des listes de littraux, c2 n'a pas de variable
  rsultat : true si et seulement si il y a une substitution s telle que s(c1) est inclus dans c2*)
  match c1 with
    | [] -> true
    | t ::f -> List.exists 
	(function l -> 
	   try let s = filtre_litteraux t l in filtre (List.map (subst_l s) f) c2
           with Echec_filtre -> false)
	c2;;

let inclus_clauses c1 c2 =
  (* rsultat : true si et seulement la clause c1 subsume la clause  c2, 
     autrement dit s'il existe une substitution s telle que s(c1) est inclus dans c2 *)
  filtre c1.corps (List.map geler_litteral c2.corps);;


(* =========================== FIN SUBSOMPTION ============================================== *)

(* ===========================================================================================
 =============================== LITTERAUX ISOLES =============================================
 ============================================================================================ *)

let unifiable l1 l2 = 
  (* rsultat : true ssi l1 est unifiable  une copie  l2 de variables disjointes de l1 *)
  let vl1 = vars_litteral l1 and vl2 = vars_litteral l2 in
  let r = renommage vl2 vl1 in 
  try let _ = unifl [l1;(subst_l r l2)] in true with Non_unifiable -> false;;

let est_isole_clause l1 c =
  (* rsultat : true ssi l'oppos du littral l1  n'est unifiable avec aucune copie des littraux de la clause c *)
  List.for_all (function l2 -> not (unifiable (opp l1) l2)) c.corps;;

let est_isole_liste_clauses l e = 
  (* rsultat : true ssi l'oppos du littral l n'est unifiable avec aucune copie des littraux des clauses de e*)
  List.for_all (function c -> est_isole_clause l c) e;;

let liste_isoles_clause c e =
  (* rsultat : la liste des littraux isols de la clause c relativement aux clauses de e *)
  let rec aux c = 
    match c with
      | [] -> []
      | tc::fc -> if est_isole_liste_clauses tc e then tc::aux fc  else aux fc 
  in aux c.corps;;

let liste_isoles e =
  (* rsultat : la liste des littraux isols des clauses de e *)
  List.fold_left (fun l1 c -> (liste_isoles_clause c (enlever c e))@l1) [] e;;






(* ========================  FIN LITTERAUX ISOLES ============================================= *)



(* Stratgie complte : identique en propositionnelle et au premier ordre *)


let rsolvants_clause_liste_de_clauses c lc =  
  (* produit la liste de tous les rsolvants de la clause  c et la liste de clauses  lc *)
  List.fold_left (fun lr d -> lr@(rsolvants c d)) [] lc;;

let rsolvants_liste_de_clauses_liste_de_clauses lc1 lc2 =
  (* produit la liste de tous les rsolvants entre les listes de clauses  lc1 et lc2 *)
  List.fold_left (fun lr c -> lr@(rsolvants_clause_liste_de_clauses c lc2)) [] lc1;;

let rec rsolvants_liste_de_clauses lc = 
  (* produit la liste de tous les rsolvants entre les clauses de lc *)
  match lc with
    | [] -> []
    | t::f -> (rsolvants_clause_liste_de_clauses t f)@(rsolvants_liste_de_clauses f);;


let limination_inclusion_clauses l  = 
  (* enlve de l les clauses  non minimales pour l'inclusion entre clauses  *)
  rduction inclus_clauses  l;;

let incluant_clauses  c l =
  (* vrifie qu'une clause  de la liste l est incluse dans la clause  c *)
  (List.exists (fun d -> inclus_clauses d c) l);;
  
let simplification_clauses l1 l2 =
  (* enlve de l1 les clauses incluants des clauses de l2 *)
  suppression  (fun c -> incluant_clauses c l2)  l1;;


(* ==========================  AFFICHAGE DES PREUVES ===================================== *)



let rec print_terme t =
  (* 1 ordre : impression du terme t*)
  match t with
    | Var x -> print_string x
    | Comp(f,l) -> print_string f; 
	if l <> [] 
	then (print_string "("; print_terme (List.hd l);
	      List.iter (fun t -> print_string ", "; print_terme t) (List.tl l);
	      print_string ")");;
      

let print_littral l  = 
  (* 1 ordre : impression du littral l *)
  match l with
    | Pos v -> print_terme v ; 
    | Neg v -> print_string "-"; print_terme v;
;;


let print_litteraux ll =
  (* impression d'une liste de littraux *)
  print_string "["; 
  if ll <> [] then
    (print_littral (List.hd ll); 
     List.iter (fun l -> print_string " "; print_littral l) (List.tl ll));
  print_string "]";;

let print_substitution s =
  (* affiche une substitution s *)
  if s <> [] then 
    let (x,t)::fs = s 
    in print_string "     "; print_string x; print_string " := ";print_terme t;
      List.iter (fun (x,t) -> print_string "; "; print_string x; print_string " := ";print_terme t; ) fs; print_string "\n";;

let print_clause_et_numrote c n =
  (* incrmente n puis attribue  la clause c le numro !n et l'affiche *)
  n := !n +1; c.numero <- !n; print_string "("; print_int !n; print_string ") "; print_litteraux c.corps;;

let print_preuve c =
  (* affiche la preuve de la clause c. Les numros des clauses servent  deux buts,
  bien prsenter la preuve et viter de rpter des preuves  *)
  let n = ref 0 (*sert  numroter les tapes de la preuve *) in
  let rec aux_print_preuve c = match c.origine with
    | Hyp -> 
	begin
	  print_clause_et_numrote c n;
	  print_string " Hyp"; print_string "\n"
	end
    | ResBin (c1,c2,s) ->
	begin
	  if c1.numero = 0 then aux_print_preuve c1;
	  if c2.numero = 0 then aux_print_preuve c2;
	  print_clause_et_numrote c n;
	  print_string " Res Bin "; print_int c1.numero; print_string ", "; print_int c2.numero; print_string "\n";
	  print_substitution s 
	end
    | Facteur (c1,s) ->
	begin
	  if c1.numero = 0 then aux_print_preuve c1;
	  print_clause_et_numrote c n;
	  print_string " Facteur "; print_int c1.numero; print_string "\n"; 
	  print_substitution s
	end
    | Copie (c1,s) ->
	begin
	  if c1.numero = 0 then aux_print_preuve c1;
	  print_clause_et_numrote c n;
	  print_string " Copie "; print_int c1.numero; print_string "\n"; 
	  print_substitution s
	end
  in aux_print_preuve c;;


  
let print_preuves f  =
  (* affiche les preuves des clauses de la liste f *)
  List.iter (fun c -> print_preuve c; print_string "\n\n") f;;



(* ============================= FIN AFFICHAGE DES PREUVES ==================== *)

(* ============================= STRATEGIE COMPLETE =========================== *)

(* Rappel : la clause c rsume la clause d, si une instance de c est incluse dans d 
il est vident que l'on peut supprimer les clauses rsumes *)

let trace = ref false;;
let trace_n = ref 0;;
let traceSc e f =
  print_string "e(";print_int !trace_n;print_string ")\n\n";
  List.iter (fun c -> print_litteraux c.corps;print_string "\n") e;
  print_string "\n\n";
  print_string "f("; print_int !trace_n;print_string ")\n\n";
  List.iter (fun c -> print_litteraux c.corps;print_string "\n") f;
  print_string "\n\n\n\n";;

(* contrle des longueurs et hauteurs des clauses *)

let max_longueur_clause = ref 5;;
let max_hauteur_clause = ref 3;;



let rec sc e f  = 
  (* sc est une ralisation rcursive terminale de la stratgie complte. 
     (sc [] f) affiche les preuves des clauses de la liste f
     (sc e f) o e <> f, appelle (sc es fs) o
     la liste de clauses es est obtenue 
     - en construisant l'ensemble lr de tous les rsolvants 
       entre les clauses de e et celles de e @ f
     - en enlevant les clauses valides, celles rsumes par une autre clause de lr, celles
       rsumes par une clause de e @ f 
     la liste de clauses fs est obtenue
     - en enlevant de e @ f les clauses rsumes par une clause de es
  *)
  if !trace then traceSc e f;
  if e = [] 
  then print_preuves f 
  else 
    let ef = e @ f in 
    let lre = rsolvants_liste_de_clauses e in
    let lref = rsolvants_liste_de_clauses_liste_de_clauses e f in
    let lr = lre @ lref in
      (* lr est la liste de tous les rsolvants entre les clauses de et celles de e  f *)
    let es = simplification_clauses (limination_inclusion_clauses (suppression_valides lr)) ef in
    let fs = simplification_clauses ef es in
      (* suppression des litteraux isols *)
    let li = liste_isoles (es@fs) in
    let esmi = suppression (fun c -> not (disjoint li c.corps)) es in
    let fsmi = suppression (fun c -> not (disjoint li c.corps)) fs in
      (* fin suppression des littraux isols *)
      (* suppression des clauses trop longues ou trop hautes *)
    let esmtrop = suppression 
		    (fun c -> ((hauteur_clause c) > !max_hauteur_clause) || 
		       ((List.length c.corps)> !max_longueur_clause)) esmi in
    let fsmtrop = suppression 
		    (fun c -> ((hauteur_clause c) > !max_hauteur_clause) || 
		       ((List.length c.corps)> !max_longueur_clause)) fsmi in
      (* fin suppression des clauses trop longues ou trop hautes *)
      (trace_n := !trace_n + 1; sc esmtrop fsmtrop);;
	 

let algo31 e =
  (* La stratgie complte est applique  la liste e de clauses et
     affiche l'ensemble des preuves des consquences minimales de e *)
  let e1 = (limination_inclusion_clauses (suppression_valides (ens_de_liste e))) in
    (* suppression des litteraux isols *)
  let li = liste_isoles e1 in
  let e1mi = suppression (fun c -> not (disjoint li c.corps)) e1 in 
    (* fin suppression des littraux isols *)
    (trace_n:=0;sc e1mi []) ;;



(* =============================== FIN STATEGIE COMPLETE ====================== *)
 

(* ============================ CONTROLE DU TEMPS D'EXECUTION ================= *)

(*  on arrte les preuves au bout de !delay secondes *)

let start = ref 0.0;;

let delay = ref 10;;

exception Too_long;;

let clock  = Sys.Signal_handle (function _ -> raise Too_long);;

let print_time () = print_string "Dure de la preuve : ";
  print_float (Unix.gettimeofday() -. !start); print_newline ();;

let algo31borne e =
  (* mthode de rsolution avec limite de temps delay *)
  try 
    (Sys.signal Sys.sigalrm clock; Unix.alarm !delay;
    start := Unix.gettimeofday();
    algo31 e;
    print_time();
    Sys.signal Sys.sigalrm Sys.Signal_ignore;())
  with 
      Too_long ->   (Sys.signal Sys.sigalrm Sys.Signal_ignore;
	print_string "Preuve interrompue (plus de ";print_int !delay;print_string " secondes)\n");;

 (* ======================= FIN CONTROLE DU TEMPS D'EXECUTION =================== *)   





(* =========================================================================================
=========================== PETITE INTERFACE ===============================================
===========================================================================================*)


let afficher_aide () =
  print_string "une commande doit tre suivie par ;;\n"; 
  print_string "les commandes sont:\n";
  print_string " quitter\n";
  print_string " aide\n"; 
  print_string " positive (pour utiliser la rsolution positive, option par dfaut)\n";
  print_string " ngative (pour utiliser la rsolution ngative)\n";
  print_string " normale (pour utiliser la rsolution normale)\n";
  print_string " trace (pour obtenir une trace de la stratgie)\n";
  print_string " sans_trace (pour supprimer la trace)\n";
  print_string " max_lc n (dfinit la longueur maximum n d'une clause)\n";
  print_string " max_hc h (dfinit la hauteur maximum h d'une clause)\n";
  print_string " max_dure t (arrte la preuve au bout de t secondes)\n";
  print_string " info (rappelle les paramtres du prouveur)\n";
  print_string " preuve suivi d'une liste de clauses (exemple : preuve a(x) b(f(x)), -a(c) b(y))\n";
  print_string " \tles clauses sont spares par des virgules\n";
  print_string " \tune clause est une suite de littraux\n";
  print_string " \tun littral positif est un atome \n";
  print_string " \tun littral ngatif est un atome prcd du signe moins\n";
  print_string " \tun atome est un identificateur suivi d'une liste de termes entre parenthses\n";
  print_string " \tune liste de termes peut tre vide et les termes sont spars par des virgules\n";
  print_string " \tun terme est une constante ou identificateur suivi\n\t\t d'une liste de termes entre parenthses\n";
  print_string " \tune constante est un identificateur commenant par  'a,b,c,d,e,f'\n\t\tou un identificateur suivi de ()\n";;
 

let afficher_info () =
  print_string "rsolution : "; 
  if !type_rsolution = pos then print_string "positive"
  else if !type_rsolution = neg then print_string "ngative"
  else print_string "normale";
  print_string "\n";
  print_string "dure maximum d'une preuve : "; print_int !delay;
  print_string "\n";
  print_string "hauteur maximum d'une clause : "; print_int !max_hauteur_clause;
  print_string "\n";
  print_string "longueur maximum d'une clause : "; print_int !max_longueur_clause;
  print_string "\n";;
    



exception Bye;;

let executer_commande c = 
  (* excute la commande c *)
  match c with
    | Quitter -> raise Bye
    | Aide -> afficher_aide (); print_newline ()
    | Negative -> type_rsolution := neg
    | Positive -> type_rsolution := pos
    | Normale -> type_rsolution := nor
    | Trace -> trace := true
    | SansTrace -> trace := false
    | Info -> afficher_info (); print_string "\n"
    | MaxDuree d -> delay := d
    | MaxLongueurClause l -> max_longueur_clause := l
    | MaxHauteurClause h -> max_hauteur_clause := h
    | Preuve l -> algo31borne l 
;;

let prompt = ">>";;

let lire_commande channel = Parser.commande Lexer.terminal (Lexing.from_channel channel);;

let interaction () =
  (* la boucle d'interaction traditionnelle *)
  try while true do
    try print_string prompt; flush stdout; executer_commande (lire_commande stdin)
    with 
      | Lexer.Erreur_lexicale ->
	  Sys.signal Sys.sigalrm Sys.Signal_ignore;
	  print_string "Erreur lexicale\n"
      | Parsing.Parse_error -> 
	  Sys.signal Sys.sigalrm Sys.Signal_ignore;
	  print_string "Erreur syntaxique\n"
  done
  with Bye -> ();;

print_string "Taper \"aide;;\" pour obtenir des informations";
print_newline ();
interaction () ;; 

 
